from enum import Enum
import ldap
from ldap.ldapobject import ReconnectLDAPObject
import srvlookup
from csh_ldap.member import CSHMember
from csh_ldap.group import CSHGroup
from csh_ldap.utility import reconnect_on_fail


class CSHLDAP:
    __domain__ = "csh.rit.edu"

    @reconnect_on_fail
    def __init__(self, bind_dn, bind_pw, *, batch_mods=False,
                 sasl=False, ro=False):
        """Handler for bindings to CSH LDAP.

        Keyword arguments:
        batch_mods -- whether or not to batch LDAP writes (default False)
        sasl -- whether or not to bypass bind_dn and bind_pw and use SASL bind
        ro -- whether or not CSH LDAP is in read only mode (default False)
        """
        if ro:
            print("########################################\n"
                  "#                                      #\n"
                  "#    CSH LDAP IS IN READ ONLY MODE     #\n"
                  "#                                      #\n"
                  "########################################")
        ldap_srvs = srvlookup.lookup("ldap", "tcp", self.__domain__)
        self.ldap_uris = ['ldaps://' + uri.hostname for uri in ldap_srvs]
        self.server_uri = None
        self.__con__ = None
        for uri in self.ldap_uris:
            try:
                self.__con__ = ReconnectLDAPObject(uri)
                self.server_uri = uri
            except (ldap.SERVER_DOWN, ldap.TIMEOUT):
                continue

        if self.__con__ is None:
            raise ldap.SERVER_DOWN

        if sasl:
            self.__con__.sasl_non_interactive_bind_s('')
        else:
            self.__con__.simple_bind_s(bind_dn, bind_pw)
        self.__mod_queue__ = {}
        self.__pending_mod_dn__ = []
        self.__batch_mods__ = batch_mods
        self.__ro__ = ro

    @reconnect_on_fail
    def get_member(self, val, uid=False):
        """Get a CSHMember object.

        Arguments:
        val -- the uuid (or uid) of the member

        Keyword arguments:
        uid -- whether or not val is a uid (default False)
        """
        return CSHMember(self, val, uid)

    @reconnect_on_fail
    def get_member_ibutton(self, val):
        """Get a CSHMember object.

        Arguments:
        val -- the iButton ID of the member

        Returns:
        None if the iButton supplied does not correspond to a CSH Member
        """
        members = self.__con__.search_s(
            CSHMember.__ldap_user_ou__,
            ldap.SCOPE_SUBTREE,
            f"(ibutton={val})",
            ['ipaUniqueID'])
        if members:
            return CSHMember(
                self,
                members[0][1]['ipaUniqueID'][0].decode('utf-8'),
                False)
        return None

    @reconnect_on_fail
    def get_member_slackuid(self, slack):
        """Get a CSHMember object.

        Arguments:
        slack -- the Slack UID of the member

        Returns:
        None if the Slack UID provided does not correspond to a CSH Member
        """
        members = self.__con__.search_s(
            CSHMember.__ldap_user_ou__,
            ldap.SCOPE_SUBTREE,
            f"(slackuid={slack})",
            ['ipaUniqueID'])
        if members:
            return CSHMember(
                self,
                members[0][1]['ipaUniqueID'][0].decode('utf-8'),
                False)
        return None

    @reconnect_on_fail
    def get_group(self, val):
        """Get a CSHGroup object.

        Arguments:
        val -- the cn of the group

        """
        return CSHGroup(self, val)

    def get_con(self):
        """Get the PyLDAP Connection"""
        return self.__con__

    @reconnect_on_fail
    def get_directorship_heads(self, val):
        """Get the head of a directorship

        Arguments:
        val -- the cn of the directorship
        """

        __ldap_group_ou__ = "cn=groups,cn=accounts,dc=csh,dc=rit,dc=edu"

        res = self.__con__.search_s(
            __ldap_group_ou__,
            ldap.SCOPE_SUBTREE,
            f"(cn=eboard-{val})",
            ['member'])

        ret = []
        for member in res[0][1]['member']:
            try:
                ret.append(member.decode('utf-8'))
            except UnicodeDecodeError:
                ret.append(member)
            except KeyError:
                continue

        return [CSHMember(self,
                          dn.split('=')[1].split(',')[0],
                          True)
                for dn in ret]

    def enqueue_mod(self, dn, mod):
        """Enqueue a LDAP modification.

        Arguments:
        dn -- the distinguished name of the object to modify
        mod -- an ldap modfication entry to enqueue
        """
        # mark for update
        if dn not in self.__pending_mod_dn__:
            self.__pending_mod_dn__.append(dn)
            self.__mod_queue__[dn] = []

        self.__mod_queue__[dn].append(mod)

    @reconnect_on_fail
    def flush_mod(self):
        """Flush all pending LDAP modifications."""
        for dn in self.__pending_mod_dn__:
            try:
                if self.__ro__:
                    for mod in self.__mod_queue__[dn]:
                        if mod[0] == ldap.MOD_DELETE:
                            mod_str = "DELETE"
                        elif mod[0] == ldap.MOD_ADD:
                            mod_str = "ADD"
                        else:
                            mod_str = "REPLACE"
                        print(f"{mod_str} VALUE {mod[1]} = {mod[2]} FOR {dn}")
                else:
                    self.__con__.modify_s(dn, self.__mod_queue__[dn])
            except ldap.TYPE_OR_VALUE_EXISTS:
                print(f"Error! Conflicting Batch Modification: {self.__mod_queue__[dn]}")
                continue
            except ldap.NO_SUCH_ATTRIBUTE:
                print(f"Error! Conflicting Batch Modification: {self.__mod_queue__[dn]}")
                continue
            self.__mod_queue__[dn] = None
        self.__pending_mod_dn__ = []



import ldap
from csh_ldap.member import CSHMember
from csh_ldap.utility import reconnect_on_fail


class CSHGroup:
    __ldap_group_ou__ = "cn=groups,cn=accounts,dc=csh,dc=rit,dc=edu"
    __ldap_base_dn__ = "dc=csh,dc=rit,dc=edu"

    @reconnect_on_fail
    def __init__(self, lib, search_val):
        """Object Model for CSH LDAP groups.

        Arguments:
        lib -- handle to a CSHLDAP instance
        search_val -- the cn of the LDAP group to bind to
        """
        self.__dict__['__lib__'] = lib
        self.__dict__['__con__'] = lib.get_con()

        res = self.__con__.search_s(
            self.__ldap_group_ou__,
            ldap.SCOPE_SUBTREE,
            f"(cn={search_val})",
            ['cn'])

        if res:
            self.__dict__['__dn__'] = res[0][0]
        else:
            raise KeyError("Invalid Search Name")

    @reconnect_on_fail
    def get_members(self):
        """Return all members in the group as CSHMember objects"""
        res = self.__con__.search_s(
            self.__ldap_base_dn__,
            ldap.SCOPE_SUBTREE,
            f"(memberof={self.__dn__})",
            ['uid'])

        ret = []
        for val in res:
            val = val[1]['uid'][0]
            try:
                ret.append(val.decode('utf-8'))
            except UnicodeDecodeError:
                ret.append(val)
            except KeyError:
                continue

        return [CSHMember(self.__lib__,
                          result,
                          uid=True)
                for result in ret]

    @reconnect_on_fail
    def check_member(self, member, dn=False):
        """Check if a Member is in the bound group.

        Arguments:
        member -- the CSHMember object (or distinguished name) of the member to
                  check against

        Keyword arguments:
        dn -- whether or not member is a distinguished name
        """

        if dn:
            res = self.__con__.search_s(
                self.__dn__,
                ldap.SCOPE_BASE,
                f"(member={dn})",
                ['ipaUniqueID'])
        else:
            res = self.__con__.search_s(
                self.__dn__,
                ldap.SCOPE_BASE,
                f"(member={member.get_dn()})",
                ['ipaUniqueID'])
        return len(res) > 0

    @reconnect_on_fail
    def add_member(self, member, dn=False):
        """Add a member to the bound group

        Arguments:
        member -- the CSHMember object (or distinguished name) of the member

        Keyword arguments:
        dn -- whether or not member is a distinguished name
        """

        if dn:
            if self.check_member(member, dn=True):
                return
            mod = (ldap.MOD_ADD, 'member', member.encode('ascii'))
        else:
            if self.check_member(member):
                return
            mod = (ldap.MOD_ADD, 'member', member.get_dn().encode('ascii'))

        if self.__lib__.__batch_mods__:
            self.__lib__.enqueue_mod(self.__dn__, mod)
        elif not self.__lib__.__ro__:
            mod_attrs = [mod]
            self.__con__.modify_s(self.__dn__, mod_attrs)
        else:
            print(f"ADD VALUE member = {mod[2]} FOR {self.__dn__}")

    @reconnect_on_fail
    def del_member(self, member, dn=False):
        """Remove a member from the bound group

        Arguments:
        member -- the CSHMember object (or distinguished name) of the member

        Keyword arguments:
        dn -- whether or not member is a distinguished name
        """

        if dn:
            if not self.check_member(member, dn=True):
                return
            mod = (ldap.MOD_DELETE, 'member', member.encode('ascii'))
        else:
            if not self.check_member(member):
                return
            mod = (ldap.MOD_DELETE, 'member', member.get_dn().encode('ascii'))

        if self.__lib__.__batch_mods__:
            self.__lib__.enqueue_mod(self.__dn__, mod)
        elif not self.__lib__.__ro__:
            mod_attrs = [mod]
            self.__con__.modify_s(self.__dn__, mod_attrs)
        else:
            print(f"DELETE VALUE member = {mod[2]} FOR {self.__dn__}")



from functools import wraps
import ldap
import srvlookup

MAX_RECONNECTS = 3


def reconnect_on_fail(method):
    """
    Decorator for CSHLDAP operations that attempts to reconnect and recall a
    method on a failed call.

    :param method: Method to call
    :return: wrapper function
    """

    @wraps(method)
    def wrapper(*method_args, **method_kwargs):
        """
        Wrapper for method, calls method and returns the result if successful,
         otherwise tries reconnecting.

        :param method_args: method's arguments
        :param method_kwargs: method's keyword arguments
        :return: result of method call
        """
        max_reconnects = MAX_RECONNECTS

        def is_cshldap(arg):
            return any(
                filter(
                    lambda t: t.__name__ == 'CSHLDAP',
                    type(arg).__mro__
                )
            )
        ldap_obj = next(filter(is_cshldap, method_args)) \
                   if any(filter(is_cshldap, method_args)) \
                   else method_args[0].__lib__
        while max_reconnects:
            try:
                result = method(*method_args, **method_kwargs)
                return result
            except (ldap.SERVER_DOWN, ldap.TIMEOUT):
                ldap_srvs = srvlookup.lookup(
                    "ldap", "tcp", ldap_obj.__domain__)
                ldap_obj.ldap_uris = ['ldaps://' + uri.hostname
                                      for uri in ldap_srvs]

                for uri in ldap_obj.ldap_uris:
                    try:
                        ldap_obj.__con__.reconnect(uri)
                        ldap_obj.server_uri = uri
                        result = method(*method_args, **method_kwargs)
                        return result
                    except (ldap.SERVER_DOWN, ldap.TIMEOUT):
                        continue
                max_reconnects -= 1
                if max_reconnects == 0:
                    raise
        return None

    return wrapper



import ldap
from csh_ldap.utility import reconnect_on_fail


class CSHMember:
    __ldap_user_ou__ = "cn=users,cn=accounts,dc=csh,dc=rit,dc=edu"
    __ldap_group_ou__ = "cn=groups,cn=accounts,dc=csh,dc=rit,dc=edu"

    @reconnect_on_fail
    def __init__(self, lib, search_val, uid):
        """Object Model for CSH LDAP users.

        Arguments:
        lib -- handle to a CSHLDAP instance
        search_val -- the uuid (or uid) of the member to bind to
        uid -- whether or not search_val is a uid
        """
        self.__dict__['__lib__'] = lib
        self.__dict__['__con__'] = lib.get_con()

        res = None

        if uid:
            res = self.__con__.search_s(
                self.__ldap_user_ou__,
                ldap.SCOPE_SUBTREE,
                f"(uid={search_val})",
                ['ipaUniqueID'])
        else:
            res = self.__con__.search_s(
                self.__ldap_user_ou__,
                ldap.SCOPE_SUBTREE,
                f"(ipaUniqueID={search_val})",
                ['uid'])

        if res:
            self.__dict__['__dn__'] = res[0][0]
        else:
            raise KeyError("Invalid Search Name")

    def __eq__(self, other):
        if isinstance(other, CSHMember):
            return self.__dn__ == other.__dn__
        return False

    def __hash__(self):
        """Generate a unique hash value for the bound CSH LDAP member object.
        """
        return hash(self.__dn__)

    def __repr__(self):
        """Generate a str representation of the bound CSH LDAP member object.
        """
        return f"CSH Member(dn: {self.__dn__})"

    def get(self, key):
        """Get an attribute from the bound CSH LDAP member object.

        Arguments:
        key -- the attribute to get the value of
        """
        return self.__getattr__(key, as_list=True) # pylint: disable=unnecessary-dunder-call

    @reconnect_on_fail
    def groups(self):
        """Get the list of Groups (by dn) that the bound CSH LDAP member object
        is in.
        """
        group_list = []
        all_groups = self.get('memberof')
        for group_dn in all_groups:
            if self.__ldap_group_ou__ in group_dn:
                group_list.append(group_dn)

        return group_list

    def in_group(self, group, dn=False):
        """Get whether or not the bound CSH LDAP member object is part of a
        group.

        Arguments:
        group -- the CSHGroup object (or distinguished name) of the group to
                 check membership for
        """
        if dn:
            return group in self.groups()
        return group.check_member(self)

    def get_dn(self):
        """Get the distinguished name of the bound LDAP object"""
        return self.__dn__

    @reconnect_on_fail
    def __getattr__(self, key, as_list=False):
        res = self.__con__.search_s(
            self.__dn__,
            ldap.SCOPE_BASE,
            "(objectClass=*)",
            [key])

        if as_list:
            ret = []
            for val in res[0][1][key]:
                try:
                    ret.append(val.decode('utf-8'))
                except UnicodeDecodeError:
                    ret.append(val)
                except KeyError:
                    continue

            return ret
        try:
            return res[0][1][key][0].decode('utf-8')
        except UnicodeDecodeError:
            return res[0][1][key][0]
        except KeyError:
            return None

    @reconnect_on_fail
    def __setattr__(self, key, value):
        ldap_mod = None

        exists = self.__con__.search_s(
            self.__dn__,
            ldap.SCOPE_BASE,
            "(objectClass=*)",
            [key])

        if value is None or value == "":
            ldap_mod = ldap.MOD_DELETE
            if exists[0][1] == {}:
                # if element doesn't exist STOP
                return
        elif exists[0][1] == {}:
            ldap_mod = ldap.MOD_ADD
        else:
            ldap_mod = ldap.MOD_REPLACE

        if value is None:
            mod = (ldap_mod, key, None)
        else:
            mod = (ldap_mod, key, value.encode('ascii'))

        if self.__lib__.__batch_mods__:
            self.__lib__.enqueue_mod(self.__dn__, mod)
        elif not self.__lib__.__ro__:
            mod_attrs = [mod]
            self.__con__.modify_s(self.__dn__, mod_attrs)
        else:
            if ldap_mod == ldap.MOD_DELETE:
                mod_str = "DELETE"
            elif ldap_mod == ldap.MOD_ADD:
                mod_str = "ADD"
            else:
                mod_str = "REPLACE"
            print(f"{mod_str} FIELD {key} WITH {value} FOR {self.__dn__}")
